from enum import Enum
from typing import TYPE_CHECKING, AbstractSet, Dict, FrozenSet, NamedTuple, Optional, Sequence

import dagster._check as check
from dagster._annotations import experimental, public
from dagster._serdes.serdes import (
    NamedTupleSerializer,
    UnpackContext,
    UnpackedValue,
    whitelist_for_serdes,
)

if TYPE_CHECKING:
    from dagster._core.definitions.auto_materialize_rule import (
        AutoMaterializeRule,
        AutoMaterializeRuleSnapshot,
    )


class AutoMaterializePolicySerializer(NamedTupleSerializer):
    def before_unpack(
        self, context: UnpackContext, unpacked_dict: Dict[str, UnpackedValue]
    ) -> Dict[str, UnpackedValue]:
        from dagster._core.definitions.auto_materialize_rule import AutoMaterializeRule

        backcompat_map = {
            "on_missing": AutoMaterializeRule.materialize_on_missing(),
            "on_new_parent_data": AutoMaterializeRule.materialize_on_parent_updated(),
            "for_freshness": AutoMaterializeRule.materialize_on_required_for_freshness(),
        }

        # determine if this namedtuple was serialized with the old format (booleans for rules)
        if any(backcompat_key in unpacked_dict for backcompat_key in backcompat_map):
            # all old policies had these rules by default
            rules = {
                AutoMaterializeRule.skip_on_parent_outdated(),
                AutoMaterializeRule.skip_on_parent_missing(),
                AutoMaterializeRule.skip_on_required_but_nonexistent_parents(),
                AutoMaterializeRule.skip_on_backfill_in_progress(),
            }
            for backcompat_key, rule in backcompat_map.items():
                if unpacked_dict.get(backcompat_key):
                    rules.add(rule)
            unpacked_dict["rules"] = frozenset(rules)

        return unpacked_dict


class AutoMaterializePolicyType(Enum):
    EAGER = "EAGER"
    LAZY = "LAZY"


@experimental
@whitelist_for_serdes(
    old_fields={"time_window_partition_scope_minutes": 1e-6},
    serializer=AutoMaterializePolicySerializer,
)
class AutoMaterializePolicy(
    NamedTuple(
        "_AutoMaterializePolicy",
        [
            ("rules", FrozenSet["AutoMaterializeRule"]),
            ("max_materializations_per_minute", Optional[int]),
        ],
    )
):
    """An AutoMaterializePolicy specifies how Dagster should attempt to keep an asset up-to-date.

    Each policy consists of a set of AutoMaterializeRules, which are used to determine whether an
    asset or a partition of an asset should or should not be auto-materialized.

    The most common policy is `AutoMaterializePolicy.eager()`, which consists of the following rules:

    - `AutoMaterializeRule.materialize_on_missing()`
        Materialize an asset or a partition if it has never been materialized.
    - `AutoMaterializeRule.materialize_on_parent_updated()`
        Materialize an asset or a partition if one of its parents have been updated more recently
        than it has.
    - `AutoMaterializeRule.materialize_on_required_for_freshness()`
        Materialize an asset or a partition if it is required to satisfy a freshness policy.
    - `AutoMaterializeRule.skip_on_parent_outdated()`
        Skip materializing an asset or partition if any of its parents have ancestors that have
        been materialized more recently.
    - `AutoMaterializeRule.skip_on_parent_missing()`
        Skip materializing an asset or a partition if any parent has never been materialized or
        observed.

    Policies can be customized by adding or removing rules. For example, if you'd like to allow
    an asset to be materialized even if some of its parent partitions are missing:

    .. code-block:: python

        from dagster import AutoMaterializePolicy, AutoMaterializeRule

        my_policy = AutoMaterializePolicy.eager().without_rules(
            AutoMaterializeRule.skip_on_parent_missing(),
        )

    If you'd like an asset to wait for all of its parents to be updated before materializing:

    .. code-block:: python

        from dagster import AutoMaterializePolicy, AutoMaterializeRule

        my_policy = AutoMaterializePolicy.eager().with_rules(
            AutoMaterializeRule.skip_on_all_parents_not_updated(),
        )

    Lastly, the `max_materializations_per_minute` parameter, which is set to 1 by default,
    rate-limits the number of auto-materializations that can occur for a particular asset within
    a short time interval. This mainly matters for partitioned assets. Its purpose is to provide a
    safeguard against "surprise backfills", where user-error causes auto-materialize to be
    accidentally triggered for large numbers of partitions at once.

    **Warning:**

    Constructing an AutoMaterializePolicy directly is not recommended as the API is subject to change.
    AutoMaterializePolicy.eager() and AutoMaterializePolicy.lazy() are the recommended API.

    """

    def __new__(
        cls,
        rules: AbstractSet["AutoMaterializeRule"],
        max_materializations_per_minute: Optional[int] = 1,
    ):
        from dagster._core.definitions.auto_materialize_rule import AutoMaterializeRule

        check.invariant(
            max_materializations_per_minute is None or max_materializations_per_minute > 0,
            "max_materializations_per_minute must be positive. To disable rate-limiting, set it"
            " to None. To disable auto materializing, remove the policy.",
        )

        return super(AutoMaterializePolicy, cls).__new__(
            cls,
            rules=frozenset(check.set_param(rules, "rules", of_type=AutoMaterializeRule)),
            max_materializations_per_minute=max_materializations_per_minute,
        )

    @property
    def materialize_rules(self) -> AbstractSet["AutoMaterializeRule"]:
        from dagster._core.definitions.auto_materialize_rule import AutoMaterializeDecisionType

        return {
            rule
            for rule in self.rules
            if rule.decision_type == AutoMaterializeDecisionType.MATERIALIZE
        }

    @property
    def skip_rules(self) -> AbstractSet["AutoMaterializeRule"]:
        from dagster._core.definitions.auto_materialize_rule import AutoMaterializeDecisionType

        return {
            rule for rule in self.rules if rule.decision_type == AutoMaterializeDecisionType.SKIP
        }

    @public
    @staticmethod
    def eager(max_materializations_per_minute: Optional[int] = 1) -> "AutoMaterializePolicy":
        """Constructs an eager AutoMaterializePolicy.

        Args:
            max_materializations_per_minute (Optional[int]): The maximum number of
                auto-materializations for this asset that may be initiated per minute. If this limit
                is exceeded, the partitions which would have been materialized will be discarded,
                and will require manual materialization in order to be updated. Defaults to 1.
        """
        from dagster._core.definitions.auto_materialize_rule import AutoMaterializeRule

        return AutoMaterializePolicy(
            rules={
                AutoMaterializeRule.materialize_on_missing(),
                AutoMaterializeRule.materialize_on_parent_updated(),
                AutoMaterializeRule.materialize_on_required_for_freshness(),
                AutoMaterializeRule.skip_on_parent_outdated(),
                AutoMaterializeRule.skip_on_parent_missing(),
                AutoMaterializeRule.skip_on_required_but_nonexistent_parents(),
                AutoMaterializeRule.skip_on_backfill_in_progress(),
            },
            max_materializations_per_minute=check.opt_int_param(
                max_materializations_per_minute, "max_materializations_per_minute"
            ),
        )

    @public
    @staticmethod
    def lazy(max_materializations_per_minute: Optional[int] = 1) -> "AutoMaterializePolicy":
        """Constructs a lazy AutoMaterializePolicy.

        Args:
            max_materializations_per_minute (Optional[int]): The maximum number of
                auto-materializations for this asset that may be initiated per minute. If this limit
                is exceeded, the partitions which would have been materialized will be discarded,
                and will require manual materialization in order to be updated. Defaults to 1.
        """
        from dagster._core.definitions.auto_materialize_rule import AutoMaterializeRule

        return AutoMaterializePolicy(
            rules={
                AutoMaterializeRule.materialize_on_required_for_freshness(),
                AutoMaterializeRule.skip_on_parent_outdated(),
                AutoMaterializeRule.skip_on_parent_missing(),
                AutoMaterializeRule.skip_on_required_but_nonexistent_parents(),
                AutoMaterializeRule.skip_on_backfill_in_progress(),
            },
            max_materializations_per_minute=check.opt_int_param(
                max_materializations_per_minute, "max_materializations_per_minute"
            ),
        )

    @public
    def without_rules(self, *rules_to_remove: "AutoMaterializeRule") -> "AutoMaterializePolicy":
        """Constructs a copy of this policy with the specified rules removed. Raises an error
        if any of the arguments are not rules in this policy.
        """
        non_matching_rules = set(rules_to_remove).difference(self.rules)
        check.param_invariant(
            not non_matching_rules,
            "rules_to_remove",
            f"Rules {[rule for rule in rules_to_remove if rule in non_matching_rules]} do not"
            " exist in this policy.",
        )
        return self._replace(
            rules=self.rules.difference(set(rules_to_remove)),
        )

    @public
    def with_rules(self, *rules_to_add: "AutoMaterializeRule") -> "AutoMaterializePolicy":
        """Constructs a copy of this policy with the specified rules added. If an instance of a
        provided rule with the same type exists on this policy, it will be replaced.
        """
        new_rule_types = {type(rule) for rule in rules_to_add}
        return self._replace(
            rules=set(rules_to_add).union(
                {rule for rule in self.rules if type(rule) not in new_rule_types}
            )
        )

    @property
    def policy_type(self) -> AutoMaterializePolicyType:
        from dagster._core.definitions.auto_materialize_rule import AutoMaterializeRule

        if AutoMaterializeRule.materialize_on_parent_updated() in self.rules:
            return AutoMaterializePolicyType.EAGER
        return AutoMaterializePolicyType.LAZY

    @property
    def rule_snapshots(self) -> Sequence["AutoMaterializeRuleSnapshot"]:
        return [rule.to_snapshot() for rule in self.rules]



from typing import AbstractSet, Iterable, NamedTuple, Optional, Sequence

from typing_extensions import Self

import dagster._check as check
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.repository_definition import SINGLETON_REPOSITORY_NAME
from dagster._serdes import create_snapshot_id, whitelist_for_serdes


class JobSubsetSelector(
    NamedTuple(
        "_JobSubsetSelector",
        [
            ("location_name", str),
            ("repository_name", str),
            ("job_name", str),
            ("op_selection", Optional[Sequence[str]]),
            ("asset_selection", Optional[AbstractSet[AssetKey]]),
            ("asset_check_selection", Optional[AbstractSet[AssetCheckKey]]),
        ],
    )
):
    """The information needed to resolve a job within a host process."""

    def __new__(
        cls,
        location_name: str,
        repository_name: str,
        job_name: str,
        op_selection: Optional[Sequence[str]],
        asset_selection: Optional[Iterable[AssetKey]] = None,
        asset_check_selection: Optional[Iterable[AssetCheckKey]] = None,
    ):
        asset_selection = set(asset_selection) if asset_selection else None
        asset_check_selection = (
            set(asset_check_selection) if asset_check_selection is not None else None
        )
        return super(JobSubsetSelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
            repository_name=check.str_param(repository_name, "repository_name"),
            job_name=check.str_param(job_name, "job_name"),
            op_selection=check.opt_nullable_sequence_param(op_selection, "op_selection", str),
            asset_selection=check.opt_nullable_set_param(
                asset_selection, "asset_selection", AssetKey
            ),
            asset_check_selection=check.opt_nullable_set_param(
                asset_check_selection, "asset_check_selection", AssetCheckKey
            ),
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "pipelineName": self.job_name,
            "solidSelection": self.op_selection,
        }

    def with_op_selection(self, op_selection: Optional[Sequence[str]]) -> Self:
        check.invariant(
            self.op_selection is None,
            f"Can not invoke with_op_selection when op_selection={self.op_selection} is"
            " already set",
        )
        return JobSubsetSelector(
            self.location_name, self.repository_name, self.job_name, op_selection
        )


@whitelist_for_serdes
class JobSelector(
    NamedTuple(
        "_JobSelector", [("location_name", str), ("repository_name", str), ("job_name", str)]
    )
):
    def __new__(
        cls,
        location_name: str,
        repository_name: Optional[str] = None,
        job_name: Optional[str] = None,
    ):
        return super(JobSelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
            repository_name=check.opt_str_param(
                repository_name,
                "repository_name",
                default=SINGLETON_REPOSITORY_NAME,
            ),
            job_name=check.str_param(
                job_name,
                "job_name",
                "Must provide job_name argument even though it is marked as optional in the "
                "function signature. repository_name, a truly optional parameter, is before "
                "that argument and actually optional. Use of keyword arguments is "
                "recommended to avoid confusion.",
            ),
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "jobName": self.job_name,
        }

    @property
    def selector_id(self):
        return create_snapshot_id(self)

    @staticmethod
    def from_graphql_input(graphql_data):
        return JobSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            job_name=graphql_data["jobName"],
        )


@whitelist_for_serdes
class RepositorySelector(
    NamedTuple("_RepositorySelector", [("location_name", str), ("repository_name", str)])
):
    def __new__(cls, location_name: str, repository_name: str):
        return super(RepositorySelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
            repository_name=check.str_param(repository_name, "repository_name"),
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
        }

    @property
    def selector_id(self):
        return create_snapshot_id(self)

    @staticmethod
    def from_graphql_input(graphql_data):
        return RepositorySelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
        )


class CodeLocationSelector(NamedTuple("_CodeLocationSelector", [("location_name", str)])):
    def __new__(cls, location_name: str):
        return super(CodeLocationSelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
        )

    def to_repository_selector(self) -> RepositorySelector:
        return RepositorySelector(
            location_name=self.location_name, repository_name=SINGLETON_REPOSITORY_NAME
        )


class ScheduleSelector(
    NamedTuple(
        "_ScheduleSelector",
        [("location_name", str), ("repository_name", str), ("schedule_name", str)],
    )
):
    def __new__(cls, location_name: str, repository_name: str, schedule_name: str):
        return super(ScheduleSelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
            repository_name=check.str_param(repository_name, "repository_name"),
            schedule_name=check.str_param(schedule_name, "schedule_name"),
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "scheduleName": self.schedule_name,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return ScheduleSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            schedule_name=graphql_data["scheduleName"],
        )


class ResourceSelector(NamedTuple):
    location_name: str
    repository_name: str
    resource_name: str

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "resourceName": self.resource_name,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return ResourceSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            resource_name=graphql_data["resourceName"],
        )


class SensorSelector(
    NamedTuple(
        "_SensorSelector", [("location_name", str), ("repository_name", str), ("sensor_name", str)]
    )
):
    def __new__(cls, location_name: str, repository_name: str, sensor_name: str):
        return super(SensorSelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
            repository_name=check.str_param(repository_name, "repository_name"),
            sensor_name=check.str_param(sensor_name, "sensor_name"),
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "sensorName": self.sensor_name,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return SensorSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            sensor_name=graphql_data["sensorName"],
        )


@whitelist_for_serdes
class InstigatorSelector(
    NamedTuple(
        "_InstigatorSelector", [("location_name", str), ("repository_name", str), ("name", str)]
    )
):
    def __new__(cls, location_name: str, repository_name: str, name: str):
        return super(InstigatorSelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
            repository_name=check.str_param(repository_name, "repository_name"),
            name=check.str_param(name, "name"),
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "name": self.name,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return InstigatorSelector(
            location_name=graphql_data["repositoryLocationName"],
            repository_name=graphql_data["repositoryName"],
            name=graphql_data["name"],
        )


class GraphSelector(
    NamedTuple(
        "_GraphSelector", [("location_name", str), ("repository_name", str), ("graph_name", str)]
    )
):
    """The information needed to resolve a graph within a host process."""

    def __new__(cls, location_name: str, repository_name: str, graph_name: str):
        return super(GraphSelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
            repository_name=check.str_param(repository_name, "repository_name"),
            graph_name=check.str_param(graph_name, "graph_name"),
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "graphName": self.graph_name,
        }


@whitelist_for_serdes
class PartitionSetSelector(
    NamedTuple(
        "_PartitionSetSelector",
        [("location_name", str), ("repository_name", str), ("partition_set_name", str)],
    )
):
    """The information needed to resolve a partition set within a host process."""

    def __new__(cls, location_name: str, repository_name: str, partition_set_name: str):
        return super(PartitionSetSelector, cls).__new__(
            cls,
            location_name=check.str_param(location_name, "location_name"),
            repository_name=check.str_param(repository_name, "repository_name"),
            partition_set_name=check.str_param(partition_set_name, "partition_set_name"),
        )

    def to_graphql_input(self):
        return {
            "repositoryLocationName": self.location_name,
            "repositoryName": self.repository_name,
            "partitionSetName": self.partition_set_name,
        }


class PartitionRangeSelector(
    NamedTuple(
        "_PartitionRangeSelector",
        [("start", str), ("end", str)],
    )
):
    """The information needed to resolve a partition range."""

    def __new__(cls, start: str, end: str):
        return super(PartitionRangeSelector, cls).__new__(
            cls,
            start=check.inst_param(start, "start", str),
            end=check.inst_param(end, "end", str),
        )

    def to_graphql_input(self):
        return {
            "start": self.start,
            "end": self.end,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return PartitionRangeSelector(
            start=graphql_data["start"],
            end=graphql_data["end"],
        )


class PartitionsSelector(
    NamedTuple(
        "_PartitionsSelector",
        [("partition_range", PartitionRangeSelector)],
    )
):
    """The information needed to define selection partitions.
    Using partition_range as property name to avoid shadowing Python 'range' builtin .
    """

    def __new__(cls, partition_range: PartitionRangeSelector):
        return super(PartitionsSelector, cls).__new__(
            cls,
            partition_range=check.inst_param(partition_range, "range", PartitionRangeSelector),
        )

    def to_graphql_input(self):
        return {
            "range": self.partition_range.to_graphql_input(),
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        return PartitionsSelector(
            partition_range=PartitionRangeSelector.from_graphql_input(graphql_data["range"])
        )


class PartitionsByAssetSelector(
    NamedTuple(
        "PartitionsByAssetSelector",
        [
            ("asset_key", AssetKey),
            ("partitions", Optional[PartitionsSelector]),
        ],
    )
):
    """The information needed to define partitions selection for a given asset key."""

    def __new__(cls, asset_key: AssetKey, partitions: Optional[PartitionsSelector] = None):
        return super(PartitionsByAssetSelector, cls).__new__(
            cls,
            asset_key=check.inst_param(asset_key, "asset_key", AssetKey),
            partitions=check.opt_inst_param(partitions, "partitions", PartitionsSelector),
        )

    def to_graphql_input(self):
        return {
            "assetKey": self.asset_key.to_graphql_input(),
            "partitions": self.partitions.to_graphql_input() if self.partitions else None,
        }

    @staticmethod
    def from_graphql_input(graphql_data):
        asset_key = graphql_data["assetKey"]
        partitions = graphql_data.get("partitions")
        return PartitionsByAssetSelector(
            asset_key=AssetKey.from_graphql_input(asset_key),
            partitions=PartitionsSelector.from_graphql_input(partitions) if partitions else None,
        )



from abc import ABC, abstractmethod
from typing import AbstractSet, Callable, List, Optional, Sequence, Set, cast

import dagster._check as check
from dagster._core.definitions import AssetCheckEvaluation, JobDefinition, NodeHandle
from dagster._core.definitions.events import AssetMaterialization, AssetObservation
from dagster._core.definitions.utils import DEFAULT_OUTPUT
from dagster._core.errors import DagsterError, DagsterInvariantViolationError
from dagster._core.events import (
    AssetObservationData,
    DagsterEvent,
    DagsterEventType,
    ExpectationResult,
    StepExpectationResultData,
    StepMaterializationData,
)
from dagster._core.execution.plan.objects import StepFailureData
from dagster._core.execution.plan.step import StepKind
from dagster._core.storage.dagster_run import DagsterRun


class ExecutionResult(ABC):
    @property
    @abstractmethod
    def job_def(self) -> JobDefinition: ...

    @property
    @abstractmethod
    def dagster_run(self) -> DagsterRun: ...

    @property
    @abstractmethod
    def all_events(self) -> Sequence[DagsterEvent]: ...

    @property
    @abstractmethod
    def run_id(self) -> str:
        """The unique identifier of the executed run."""
        ...

    @property
    def success(self) -> bool:
        """bool: Whether execution was successful."""
        return self.dagster_run.is_success

    @property
    def all_node_events(self) -> Sequence[DagsterEvent]:
        """List[DagsterEvent]: All dagster events from the execution."""
        step_events: List[DagsterEvent] = []

        for node_name in self.job_def.graph.node_dict.keys():
            handle = NodeHandle.from_string(node_name)
            step_events += self._filter_events_by_handle(handle)

        return step_events

    @abstractmethod
    def _get_output_for_handle(self, handle: NodeHandle, output_name: str) -> object:
        raise NotImplementedError()

    def _filter_events_by_handle(self, handle: NodeHandle) -> Sequence[DagsterEvent]:
        def _is_event_from_node(event: DagsterEvent) -> bool:
            if not event.is_step_event:
                return False
            node_handle = cast(NodeHandle, event.node_handle)
            return node_handle.is_or_descends_from(handle)

        return self.filter_events(_is_event_from_node)

    def output_value(self, output_name: str = DEFAULT_OUTPUT) -> object:
        check.str_param(output_name, "output_name")

        graph_def = self.job_def.graph
        if not graph_def.has_output(output_name) and len(graph_def.output_mappings) == 0:
            raise DagsterInvariantViolationError(
                f"Attempted to retrieve top-level outputs for '{graph_def.name}', which has no"
                " outputs."
            )
        elif not graph_def.has_output(output_name):
            raise DagsterInvariantViolationError(
                f"Could not find top-level output '{output_name}' in '{graph_def.name}'."
            )
        # Resolve the first layer of mapping
        output_mapping = graph_def.get_output_mapping(output_name)
        mapped_node = graph_def.node_named(output_mapping.maps_from.node_name)
        origin_output_def, origin_handle = mapped_node.definition.resolve_output_to_origin(
            output_mapping.maps_from.output_name,
            NodeHandle(mapped_node.name, None),
        )

        # Get output from origin node
        return self._get_output_for_handle(check.not_none(origin_handle), origin_output_def.name)

    def output_for_node(self, node_str: str, output_name: str = DEFAULT_OUTPUT) -> object:
        # resolve handle of node that node_str is referring to
        target_handle = NodeHandle.from_string(node_str)
        target_node_def = self.job_def.graph.get_node(target_handle).definition
        origin_output_def, origin_handle = target_node_def.resolve_output_to_origin(
            output_name, NodeHandle.from_string(node_str)
        )

        # retrieve output value from resolved handle
        return self._get_output_for_handle(check.not_none(origin_handle), origin_output_def.name)

    def filter_events(self, event_filter: Callable[[DagsterEvent], bool]) -> Sequence[DagsterEvent]:
        return [event for event in self.all_events if event_filter(event)]

    def events_for_node(self, node_name: str) -> Sequence[DagsterEvent]:
        """Retrieves all dagster events for a specific node.

        Args:
            node_name (str): The name of the node for which outputs should be retrieved.

        Returns:
            List[DagsterEvent]: A list of all dagster events associated with provided node name.
        """
        check.str_param(node_name, "node_name")

        return self._filter_events_by_handle(NodeHandle.from_string(node_name))

    def is_node_success(self, node_str: str) -> bool:
        return any(evt.is_step_success for evt in self.events_for_node(node_str))

    def is_node_failed(self, node_str: str) -> bool:
        return any(evt.is_step_failure for evt in self.events_for_node(node_str))

    def is_node_skipped(self, node_str: str) -> bool:
        return any(evt.is_step_skipped for evt in self.events_for_node(node_str))

    def is_node_untouched(self, node_str: str) -> bool:
        return len(self.events_for_node(node_str)) == 0

    def get_job_failure_event(self) -> DagsterEvent:
        """Returns a DagsterEvent with type DagsterEventType.PIPELINE_FAILURE if it ocurred during
        execution.
        """
        events = self.filter_events(
            lambda event: event.event_type == DagsterEventType.PIPELINE_FAILURE
        )

        if len(events) == 0:
            raise DagsterError("No event of type DagsterEventType.PIPELINE_FAILURE found.")

        return events[0]

    def get_job_success_event(self) -> DagsterEvent:
        """Returns a DagsterEvent with type DagsterEventType.PIPELINE_SUCCESS if it ocurred during
        execution.
        """
        events = self.filter_events(
            lambda event: event.event_type == DagsterEventType.PIPELINE_SUCCESS
        )

        if len(events) == 0:
            raise DagsterError("No event of type DagsterEventType.PIPELINE_SUCCESS found.")

        return events[0]

    def asset_materializations_for_node(self, node_name: str) -> Sequence[AssetMaterialization]:
        return [
            cast(StepMaterializationData, event.event_specific_data).materialization
            for event in self.events_for_node(node_name)
            if event.event_type_value == DagsterEventType.ASSET_MATERIALIZATION.value
        ]

    def asset_observations_for_node(self, node_name: str) -> Sequence[AssetObservation]:
        return [
            cast(AssetObservationData, event.event_specific_data).asset_observation
            for event in self.events_for_node(node_name)
            if event.event_type_value == DagsterEventType.ASSET_OBSERVATION.value
        ]

    def get_asset_materialization_events(self) -> Sequence[DagsterEvent]:
        return [event for event in self.all_events if event.is_step_materialization]

    def get_asset_observation_events(self) -> Sequence[DagsterEvent]:
        return [event for event in self.all_events if event.is_asset_observation]

    def get_asset_check_evaluations(self) -> Sequence[AssetCheckEvaluation]:
        return [
            cast(AssetCheckEvaluation, event.event_specific_data)
            for event in self.all_events
            if event.event_type_value == DagsterEventType.ASSET_CHECK_EVALUATION.value
        ]

    def get_step_success_events(self) -> Sequence[DagsterEvent]:
        return [event for event in self.all_events if event.is_step_success]

    def get_step_skipped_events(self) -> Sequence[DagsterEvent]:
        return [event for event in self.all_events if event.is_step_skipped]

    def get_step_failure_events(self) -> Sequence[DagsterEvent]:
        return [event for event in self.all_events if event.is_step_failure]

    def get_failed_step_keys(self) -> AbstractSet[str]:
        failure_events = self.filter_events(
            lambda event: event.is_step_failure or event.is_resource_init_failure
        )
        keys: Set[str] = set()
        for event in failure_events:
            if event.step_key:
                keys.add(event.step_key)
        return keys

    def compute_events_for_handle(self, handle: NodeHandle) -> Sequence[DagsterEvent]:
        return [
            event
            for event in self._filter_events_by_handle(handle)
            if event.step_kind == StepKind.COMPUTE
        ]

    def expectation_results_for_node(self, node_str: str) -> Sequence[ExpectationResult]:
        handle = NodeHandle.from_string(node_str)
        compute_events = self.compute_events_for_handle(handle)
        expectation_result_events = list(
            filter(
                lambda event: event.event_type == DagsterEventType.STEP_EXPECTATION_RESULT,
                compute_events,
            )
        )
        return [
            cast(StepExpectationResultData, event.event_specific_data).expectation_result
            for event in expectation_result_events
        ]

    def retry_attempts_for_node(self, node_str: str) -> int:
        count = 0
        for event in self.events_for_node(node_str):
            if event.event_type == DagsterEventType.STEP_RESTARTED:
                count += 1
        return count

    def failure_data_for_node(self, node_str: str) -> Optional[StepFailureData]:
        for event in self.events_for_node(node_str):
            if event.event_type == DagsterEventType.STEP_FAILURE:
                return event.step_failure_data
        return None
